from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import subprocess

__all__ = [
    "CMakeAccessError",
    "CMakeConfigError",
    "CMakeNotFoundError",
    "CMakeVersionError",
    "FailedLiveProcessError",
    "FailedProcessError",
    "NinjaNotFoundError",
    "NinjaVersionError",
    "NotFoundError",
    "ScikitBuildError",
]


def __dir__() -> list[str]:
    return __all__


class ScikitBuildError(Exception):
    """
    Base class for all ScikitBuildError errors.
    """


class NotFoundError(ScikitBuildError):
    """
    Raised when a program is not found.
    """


class CMakeNotFoundError(NotFoundError):
    """
    Raised when cmake is not found.
    """


class NinjaNotFoundError(NotFoundError):
    """
    Raised when ninja is not found.
    """


class FailedProcessError(Exception):
    """
    Exception raised when an call fails.
    """

    def __init__(
        self, exception: subprocess.CalledProcessError, description: str
    ) -> None:
        super().__init__()
        self.exception = exception
        self._description = description

    def __str__(self) -> str:
        cmd = " ".join(self.exception.cmd)
        description = f"{self._description}\n  Command {cmd!r} failed with return code {self.exception.returncode}"
        for stream_name in ("stdout", "stderr"):
            stream = getattr(self.exception, stream_name)
            if stream:
                description += f"\n  {stream_name}:\n"
                description += textwrap.indent(stream.decode(), "    ")
        return description


class FailedLiveProcessError(Exception):
    """
    Exception for when output was not being redirected.
    """

    def __init__(self, *args: object, msg: str = "") -> None:
        super().__init__(*args)
        self.msg = msg


class CMakeAccessError(FailedProcessError):
    """
    Error raised when CMake access fails.
    """


class CMakeVersionError(ScikitBuildError):
    """
    Error raised when CMake version is not supported.
    """


class NinjaVersionError(ScikitBuildError):
    """
    Error raised when CMake version is not supported.
    """


class CMakeConfigError(ScikitBuildError):
    """
    Something is misconfigured.
    """



# SPDX-License-Identifier: MIT

"""
This module focuses on reading pyproject.toml fields with error collection. It is
mostly internal, except for License and Readme classes, which are re-exported in
the top-level package.

These functions return None if an error occurred, and use the ErrorCollector to
record errors.
"""

from __future__ import annotations

import dataclasses
import typing
from typing import Any

import packaging.requirements

if typing.TYPE_CHECKING:
    import pathlib
    from collections.abc import Generator, Iterable

    from packaging.requirements import Requirement

    from .errors import ErrorCollector
    from .project_table import ProjectTable


__all__ = [
    "License",
    "Readme",
]


def __dir__() -> list[str]:
    return __all__


@dataclasses.dataclass(frozen=True)
class License:
    """
    This represents a classic license, which contains text, and optionally a
    file path. Modern licenses are just SPDX identifiers, which are strings.
    """

    text: str
    file: pathlib.Path | None


@dataclasses.dataclass(frozen=True)
class Readme:
    """
    This represents a readme, which contains text and a content type, and
    optionally a file path.
    """

    text: str
    file: pathlib.Path | None
    content_type: str


def ensure_str(value: object) -> str | None:
    """Ensure that a value is a string."""
    if isinstance(value, str):
        return value
    return None


def ensure_list(val: object) -> list[str] | None:
    """Ensure that a value is a list of strings."""
    if val is None:
        return None
    if not isinstance(val, list):
        return None
    for item in val:
        if not isinstance(item, str):
            return None

    return val


def ensure_dict(val: object) -> dict[str, str] | None:
    """Ensure that a value is a dictionary of strings."""
    if not isinstance(val, dict):
        return None
    for item in val.values():
        if not isinstance(item, str):
            return None
    return val


def ensure_people(val: object) -> list[tuple[str, str | None]]:
    """Ensure that a value is a list of tables with optional "name" and "email" keys."""
    if not isinstance(val, list):
        return []
    for each in val:
        if not isinstance(each, dict):
            return []
        for value in each.values():
            if not isinstance(value, str):
                return []
    return [(entry.get("name", "Unknown"), entry.get("email")) for entry in val]


def get_license(
    project: ProjectTable, project_dir: pathlib.Path, error_collector: ErrorCollector
) -> License | str | None:
    """Get the license field from the project table. Handles PEP 639 style license too.

    None is returned if the license field is not present or if an error occurred.
    """
    val = project.get("license")
    if val is None:
        return None
    if isinstance(val, str):
        return val

    if isinstance(val, dict):
        _license = ensure_dict(val)
        if _license is None:
            return None
    else:
        return None

    file: pathlib.Path | None = None
    filename = _license.get("file")
    text = _license.get("text")

    if (filename and text) or (not filename and not text):
        return None

    if filename:
        file = project_dir.joinpath(filename)
        if not file.is_file():
            msg = f"License file not found ({filename!r})"
            error_collector.config_error(msg, key="project.license.file")
            return None
        text = file.read_text(encoding="utf-8")

    assert text is not None
    return License(text, file)


def get_license_files(
    project: dict[str, Any], project_dir: pathlib.Path, error_collector: ErrorCollector
) -> list[pathlib.Path] | None:
    """Get the license-files list of files from the project table.

    Returns None if an error occurred (including invalid globs, etc) or if
    not present.
    """
    license_files = project.get("license-files")
    if license_files is None:
        return None
    if ensure_list(license_files) is None:
        return None

    return list(_get_files_from_globs(project_dir, license_files, error_collector))


def get_readme(  # noqa: C901
    project: dict[str, Any], project_dir: pathlib.Path, error_collector: ErrorCollector
) -> Readme | None:
    """Get the text of the readme from the project table.

    Returns None if an error occurred or if the readme field is not present.
    """
    if "readme" not in project:
        return None

    filename: str | None = None
    file: pathlib.Path | None = None
    text: str | None = None
    content_type: str | None = None

    readme = project["readme"]
    if isinstance(readme, str):
        # readme is a file
        text = None
        filename = readme
        if filename.endswith(".md"):
            content_type = "text/markdown"
        elif filename.endswith(".rst"):
            content_type = "text/x-rst"
        else:
            msg = "Could not infer content type for readme file {filename!r}"
            error_collector.config_error(msg, key="project.readme", filename=filename)
            return None
    elif isinstance(readme, dict):
        # readme is a dict containing either 'file' or 'text', and content-type
        for field in readme:
            if field not in ("content-type", "file", "text"):
                return None

        content_type_raw = readme.get("content-type")
        if content_type_raw is not None:
            content_type = ensure_str(content_type_raw)
            if content_type is None:
                return None
        filename_raw = readme.get("file")
        if filename_raw is not None:
            filename = ensure_str(filename_raw)
            if filename is None:
                return None

        text_raw = readme.get("text")
        if text_raw is not None:
            text = ensure_str(text_raw)
            if text is None:
                return None

        if (filename and text) or (not filename and not text):
            return None
        if not content_type:
            return None
    else:
        return None

    if filename:
        file = project_dir.joinpath(filename)
        if not file.is_file():
            msg = "Readme file not found ({filename!r})"
            error_collector.config_error(
                msg, key="project.readme.file", filename=filename
            )
            return None
        text = file.read_text(encoding="utf-8")

    assert text is not None
    return Readme(text, file, content_type)


def get_dependencies(project: dict[str, Any]) -> list[Requirement]:
    """Get the dependencies from the project table."""
    requirement_strings: list[str] | None = None
    requirement_strings_raw = project.get("dependencies")
    if requirement_strings_raw is not None:
        requirement_strings = ensure_list(requirement_strings_raw)
    if requirement_strings is None:
        return []

    requirements: list[Requirement] = []
    try:
        requirements.extend(
            packaging.requirements.Requirement(req) for req in requirement_strings
        )
    except packaging.requirements.InvalidRequirement:
        return []
    return requirements


def get_optional_dependencies(
    project: dict[str, Any],
) -> dict[str, list[Requirement]]:
    """Get the optional dependencies from the project table."""
    val = project.get("optional-dependencies")
    if not val:
        return {}

    requirements_dict: dict[str, list[Requirement]] = {}
    if not isinstance(val, dict):
        return {}
    for extra, requirements in val.copy().items():
        assert isinstance(extra, str)
        if not isinstance(requirements, list):
            return {}
        requirements_dict[extra] = []
        for req in requirements:
            if not isinstance(req, str):
                return {}
            try:
                requirements_dict[extra].append(packaging.requirements.Requirement(req))
            except packaging.requirements.InvalidRequirement:
                return {}
    return dict(requirements_dict)


def get_entrypoints(project: dict[str, Any]) -> dict[str, dict[str, str]]:
    """Get the entrypoints from the project table."""
    val = project.get("entry-points")
    if val is None:
        return {}
    if not isinstance(val, dict):
        return {}
    for entrypoints in val.values():
        if not isinstance(entrypoints, dict):
            return {}
        for entrypoint in entrypoints.values():
            if not isinstance(entrypoint, str):
                return {}
    return val


def get_dynamic(project: dict[str, Any]) -> list[str]:
    """Get the dynamic fields from the project table.

    Returns an empty list if the field is not present or if an error occurred.
    """
    return ensure_list(project.get("dynamic", [])) or []


def _get_files_from_globs(
    project_dir: pathlib.Path, globs: Iterable[str], error_collector: ErrorCollector
) -> Generator[pathlib.Path, None, None]:
    """Given a list of globs, get files that match."""
    for glob in globs:
        if glob.startswith(("..", "/")):
            msg = "{glob!r} is an invalid {key} glob: the pattern must match files within the project directory"
            error_collector.config_error(msg, key="project.license-files", glob=glob)
            break
        files = [f for f in project_dir.glob(glob) if f.is_file()]
        if not files:
            msg = "Every pattern in {key} must match at least one file: {glob!r} did not match any"
            error_collector.config_error(msg, key="project.license-files", glob=glob)
            break
        for f in files:
            yield f.relative_to(project_dir)



from __future__ import annotations

from ..ast.ast import parse
from ..ast.tokenizer import tokenize

__all__ = ["find_min_cmake_version"]


def __dir__() -> list[str]:
    return __all__


def find_min_cmake_version(cmake_content: str) -> str | None:
    """
    Locate the minimum required version. Return None if not found.
    """
    for node in parse(tokenize(cmake_content)):
        if node.name == "cmake_minimum_required":
            return (
                node.value.replace("VERSION", "")
                .replace("FATAL_ERROR", "")
                .split("...")[0]
                .strip()
                .strip("\"'[]=")
            )

    return None



from __future__ import annotations

import dataclasses
import importlib
import inspect
import sys
from collections.abc import Iterator, Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, Union, runtime_checkable

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    StrMapping = Mapping[str, Any]
else:
    StrMapping = Mapping

from ..metadata import _ALL_FIELDS

__all__ = ["load_provider", "process_dynamic_metadata"]


def __dir__() -> list[str]:
    return __all__


@runtime_checkable
class DynamicMetadataProtocol(Protocol):
    def dynamic_metadata(
        self,
        fields: Iterable[str],
        settings: dict[str, Any],
        project: Mapping[str, Any],
    ) -> dict[str, Any]: ...


@runtime_checkable
class DynamicMetadataRequirementsProtocol(DynamicMetadataProtocol, Protocol):
    def get_requires_for_dynamic_metadata(
        self, settings: dict[str, Any]
    ) -> list[str]: ...


@runtime_checkable
class DynamicMetadataWheelProtocol(DynamicMetadataProtocol, Protocol):
    def dynamic_wheel(self, field: str, settings: Mapping[str, Any]) -> bool: ...


DMProtocols = Union[
    DynamicMetadataProtocol,
    DynamicMetadataRequirementsProtocol,
    DynamicMetadataWheelProtocol,
]


def load_provider(
    provider: str,
    provider_path: str | None = None,
) -> DMProtocols:
    if provider_path is None:
        return importlib.import_module(provider)

    if not Path(provider_path).is_dir():
        msg = "provider-path must be an existing directory"
        raise AssertionError(msg)

    try:
        sys.path.insert(0, provider_path)
        return importlib.import_module(provider)
    finally:
        sys.path.pop(0)


def _load_dynamic_metadata(
    metadata: Mapping[str, Mapping[str, str]],
) -> Generator[tuple[str, DMProtocols, dict[str, Any]], None, None]:
    for field, orig_config in metadata.items():
        if "provider" not in orig_config:
            msg = "Missing provider in dynamic metadata"
            raise KeyError(msg)

        if field not in _ALL_FIELDS:
            msg = f"{field} is not a valid field"
            raise KeyError(msg)
        config = dict(orig_config)
        provider = config.pop("provider")
        provider_path = config.pop("provider-path", None)
        loaded_provider = load_provider(provider, provider_path)
        yield field, loaded_provider, config


@dataclasses.dataclass
class DynamicPyProject(StrMapping):
    settings: dict[str, dict[str, Any]]
    project: dict[str, Any]
    providers: dict[str, DMProtocols]

    def __getitem__(self, key: str) -> Any:
        # Try to get the settings from either the static file or dynamic metadata provider
        if key in self.project:
            return self.project[key]

        # Check if we are in a loop, i.e. something else is already requesting
        # this key while trying to get another key
        if key not in self.providers:
            dep_type = "missing" if key in self.settings else "circular"
            msg = f"Encountered a {dep_type} dependency at {key}"
            raise ValueError(msg)

        provider = self.providers.pop(key)
        sig = inspect.signature(provider.dynamic_metadata)
        if len(sig.parameters) < 3:
            # Backcompat for dynamic_metadata without metadata dict
            self.project[key] = provider.dynamic_metadata(  # type: ignore[call-arg]
                key, self.settings[key]
            )
        else:
            self.project[key] = provider.dynamic_metadata(key, self.settings[key], self)
        self.project["dynamic"].remove(key)

        return self.project[key]

    def __iter__(self) -> Iterator[str]:
        # Iterate over the keys of the static settings
        yield from self.project

        # Iterate over the keys of the dynamic metadata providers
        # GraalPy needs it to be a copy
        yield from list(self.providers)

    def __len__(self) -> int:
        return len(self.project) + len(self.providers)

    def __contains__(self, key: object) -> bool:
        return key in self.project or key in self.providers


def process_dynamic_metadata(
    project: Mapping[str, Any],
    metadata: Mapping[str, Mapping[str, Any]],
) -> dict[str, Any]:
    initial = {f: (p, c) for (f, p, c) in _load_dynamic_metadata(metadata)}

    settings = DynamicPyProject(
        settings={f: c for f, (_, c) in initial.items()},
        project=dict(project),
        providers={k: v for k, (v, _) in initial.items()},
    )

    return dict(settings)



from __future__ import annotations

import sys

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup
else:
    from builtins import ExceptionGroup

__all__ = ["ExceptionGroup"]


def __dir__() -> list[str]:
    return __all__



# pylint: disable=duplicate-code

from __future__ import annotations

import setuptools.build_meta
from setuptools.build_meta import (
    build_sdist,
    build_wheel,
    prepare_metadata_for_build_wheel,
)

from ..builder.get_requires import GetRequires

if hasattr(setuptools.build_meta, "build_editable"):
    from setuptools.build_meta import build_editable

if hasattr(setuptools.build_meta, "prepare_metadata_for_build_editable"):
    from setuptools.build_meta import (
        prepare_metadata_for_build_editable,
    )


__all__ = [
    "build_editable",
    "build_sdist",
    "build_wheel",
    "get_requires_for_build_editable",
    "get_requires_for_build_sdist",
    "get_requires_for_build_wheel",
    "prepare_metadata_for_build_editable",
    "prepare_metadata_for_build_wheel",
]


def __dir__() -> list[str]:
    return __all__


def get_requires_for_build_sdist(
    config_settings: dict[str, str | list[str]] | None = None,
) -> list[str]:
    setuptools_reqs = setuptools.build_meta.get_requires_for_build_sdist(
        config_settings
    )
    requires = GetRequires.from_config_settings(config_settings)

    # These are only injected if cmake is required for the SDist step
    cmake_requires = (
        [*requires.cmake(), *requires.ninja()] if requires.settings.sdist.cmake else []
    )
    return [*setuptools_reqs, *cmake_requires]


def get_requires_for_build_wheel(
    config_settings: dict[str, str | list[str]] | None = None,
) -> list[str]:
    requires = GetRequires.from_config_settings(config_settings)

    setuptools_reqs = setuptools.build_meta.get_requires_for_build_wheel(
        config_settings
    )

    return [*setuptools_reqs, *requires.cmake(), *requires.ninja()]


if hasattr(setuptools.build_meta, "get_requires_for_build_editable"):

    def get_requires_for_build_editable(
        config_settings: dict[str, str | list[str]] | None = None,
    ) -> list[str]:
        requires = GetRequires.from_config_settings(config_settings)
        setuptools_reqs = setuptools.build_meta.get_requires_for_build_editable(
            config_settings
        )
        return [*setuptools_reqs, *requires.cmake(), *requires.ninja()]



from __future__ import annotations

import copy
import sys
from typing import TYPE_CHECKING, Any

from packaging.version import Version

from .._logging import logger
from .._vendor.pyproject_metadata import (
    StandardMetadata,
    errors,
    extras_build_system,
    extras_top_level,
)
from ..builder._load_provider import process_dynamic_metadata

if TYPE_CHECKING:
    from collections.abc import Mapping

    from ..settings.skbuild_model import ScikitBuildSettings

__all__ = ["get_standard_metadata"]


def __dir__() -> list[str]:
    return __all__


# Use exceptiongroup backport
if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup

    errors.ExceptionGroup = ExceptionGroup  # type: ignore[misc, assignment]


def get_standard_metadata(
    pyproject_dict: Mapping[str, Any],
    settings: ScikitBuildSettings,
) -> StandardMetadata:
    new_pyproject_dict = copy.deepcopy(dict(pyproject_dict))

    # Handle any dynamic metadata
    new_pyproject_dict["project"] = process_dynamic_metadata(
        new_pyproject_dict["project"], settings.metadata
    )

    if settings.strict_config:
        extra_keys_top = extras_top_level(new_pyproject_dict)
        if extra_keys_top:
            logger.warning(
                f"Unknown keys in top-level of pyproject.toml: {', '.join(extra_keys_top)}"
            )
        extra_keys_build = extras_build_system(new_pyproject_dict)
        if extra_keys_build:
            logger.warning(
                f"Unknown keys in build-system of pyproject.toml: {', '.join(extra_keys_build)}"
            )

    extra_validate = (
        settings.minimum_version is None or settings.minimum_version >= Version("0.10")
    )
    if extra_validate:
        allow_extra_keys: bool | None = not settings.strict_config
    else:
        allow_extra_keys = None if settings.strict_config else False

    metadata = StandardMetadata.from_pyproject(
        new_pyproject_dict, all_errors=True, allow_extra_keys=allow_extra_keys
    )

    # For scikit-build-core < 0.5, we keep the normalized name for back-compat
    if settings.minimum_version is not None and settings.minimum_version < Version(
        "0.5"
    ):
        metadata.name = metadata.canonical_name

    # The description field is required to be one line. Instead of merging it
    # or cutting off subsequent lines (setuptools), we throw a nice error.
    # But we didn't validate before 0.9.
    if (
        settings.minimum_version is None or settings.minimum_version >= Version("0.9")
    ) and "\n" in (metadata.description or ""):
        msg = "Multiple lines in project.description are not supported; this is supposed to be a one line summary"
        raise ValueError(msg)

    # Validate license if possible.
    if isinstance(metadata.license, str):
        try:
            import packaging.licenses

            metadata.license = packaging.licenses.canonicalize_license_expression(
                metadata.license
            )
        except ImportError:
            logger.warning(
                "Packaging 24.2+ required for license normalization. Please update (Python 3.8+ required)"
            )

    # For scikit-build-core >= 0.11, we set METADATA 2.2 as minimum
    if (
        settings.minimum_version is None or settings.minimum_version >= Version("0.11")
    ) and metadata.auto_metadata_version == "2.1":
        metadata.metadata_version = "2.2"

    return metadata



"""Format variables available in the ``pyproject.toml`` evaluation"""

from __future__ import annotations

import dataclasses
import sys
import typing
from pathlib import Path
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import Literal

    from scikit_build_core.builder.wheel_tag import WheelTag
    from scikit_build_core.settings.skbuild_model import ScikitBuildSettings

__all__ = [
    "PyprojectFormatter",
    "RootPathResolver",
    "pyproject_format",
]


def __dir__() -> list[str]:
    return __all__


class PyprojectFormatter(TypedDict, total=False):
    """Format helper for pyproject.toml.

    Stores all known variables that can be used for evaluating a formatted string
    in the pyproject.toml config file.
    """

    cache_tag: str
    """Tag used by the import machinery in the filenames of cached modules, i.e. ``sys.implementation.cache_tag``."""
    wheel_tag: str
    """The tags as computed for the wheel."""
    build_type: str
    """Build type passed as ``cmake.build_type``."""
    state: Literal["sdist", "wheel", "editable", "metadata_wheel", "metadata_editable"]
    """The state of the build."""
    root: RootPathResolver
    """Root path of the current project."""


@typing.overload
def pyproject_format(
    *,
    settings: ScikitBuildSettings,
    state: Literal["sdist", "wheel", "editable", "metadata_wheel", "metadata_editable"]
    | None = ...,
    tags: WheelTag | None = ...,
) -> PyprojectFormatter: ...


@typing.overload
def pyproject_format(*, dummy: Literal[True]) -> dict[str, str]: ...


def pyproject_format(
    *,
    settings: ScikitBuildSettings | None = None,
    state: (
        Literal["sdist", "wheel", "editable", "metadata_wheel", "metadata_editable"]
        | None
    ) = None,
    tags: WheelTag | None = None,
    dummy: bool = False,
) -> PyprojectFormatter | dict[str, str]:
    """Generate :py:class:`PyprojectFormatter` dictionary to use in f-string format."""
    if dummy:
        # Return a dict with all the known keys but with values replaced with dummy values
        return dict.fromkeys(PyprojectFormatter.__annotations__, "*")

    assert settings is not None
    # First set all known values
    res = PyprojectFormatter(
        cache_tag=sys.implementation.cache_tag,
        # We are assuming the Path.cwd always evaluates to the folder containing pyproject.toml
        # as part of PEP517 standard.
        root=RootPathResolver(),
        build_type=settings.cmake.build_type,
    )
    # Then compute all optional keys depending on the function input
    if tags is not None:
        res["wheel_tag"] = str(tags)
    if state is not None:
        res["state"] = state
    # Construct the final dict including the always known keys
    return res


@dataclasses.dataclass()
class RootPathResolver:
    """Handle ``{root:uri}`` like formatting similar to ``hatchling``."""

    path: Path = dataclasses.field(default_factory=Path)

    def __post_init__(self) -> None:
        self.path = self.path.resolve()

    def __format__(self, fmt: str) -> str:
        command, _, rest = fmt.partition(":")
        if command == "parent":
            parent = RootPathResolver(self.path.parent)
            return parent.__format__(rest)
        if command == "uri" and rest == "":
            return self.path.as_uri()
        if command == "" and rest == "":
            return str(self)
        msg = f"Could not handle format: {fmt}"
        raise ValueError(msg)



from __future__ import annotations

from typing import Any

from hatchling.plugin import hookimpl

from .plugin import ScikitBuildHook

__all__ = ["hatch_register_build_hook"]


def __dir__() -> list[str]:
    return __all__


@hookimpl
def hatch_register_build_hook() -> Any:
    return ScikitBuildHook



from __future__ import annotations

import contextlib
import os
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import pathspec

from .._logging import logger
from ..format import pyproject_format

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence

__all__ = ["each_unignored_file"]

EXCLUDE_LINES = [
    ".git",
    ".tox/",
    ".nox/",
    ".egg-info/",
    "__pycache__/",
    "__pypackages__/",
    "*.pyc",
    "*.dist-info/",
]


def __dir__() -> list[str]:
    return __all__


def each_unignored_file(
    starting_path: Path,
    include: Sequence[str] = (),
    exclude: Sequence[str] = (),
    build_dir: str = "",
    *,
    mode: Literal["classic", "default", "manual"],
) -> Generator[Path, None, None]:
    """
    Runs through all non-ignored files. Must be run from the root directory.
    """
    global_exclude_lines = []
    if mode != "manual":
        for gi in [Path(".git/info/exclude"), Path(".gitignore")]:
            ignore_errs = [FileNotFoundError, NotADirectoryError]
            with contextlib.suppress(*ignore_errs), gi.open(encoding="utf-8") as f:
                global_exclude_lines += f.readlines()

    nested_excludes = (
        {}
        if mode == "manual"
        else {
            Path(dirpath): pathspec.GitIgnoreSpec.from_lines(
                (Path(dirpath) / filename).read_text(encoding="utf-8").splitlines()
            )
            for dirpath, _, filenames in os.walk(".")
            for filename in filenames
            if filename == ".gitignore" and dirpath != "."
        }
    )

    exclude_build_dir = build_dir.format(**pyproject_format(dummy=True))

    exclude_lines = (
        [*EXCLUDE_LINES, exclude_build_dir] if exclude_build_dir else EXCLUDE_LINES
    )

    user_exclude_spec = pathspec.GitIgnoreSpec.from_lines(list(exclude))
    global_exclude_spec = pathspec.GitIgnoreSpec.from_lines(global_exclude_lines)
    builtin_exclude_spec = pathspec.GitIgnoreSpec.from_lines(exclude_lines)

    include_spec = pathspec.GitIgnoreSpec.from_lines(include)

    for dirstr, dirs, filenames in os.walk(str(starting_path), followlinks=True):
        dirpath = Path(dirstr)
        if mode != "classic":
            for dname in dirs:
                if not match_path(
                    dirpath,
                    dirpath / dname,
                    include_spec,
                    global_exclude_spec,
                    builtin_exclude_spec,
                    user_exclude_spec,
                    nested_excludes,
                    is_path=True,
                ):
                    # Check to see if any include rules start with this
                    dstr = (dirpath / dname).as_posix().strip("/") + "/"
                    if not any(p.lstrip("/").startswith(dstr) for p in include):
                        dirs.remove(dname)

        for fn in filenames:
            path = dirpath / fn
            if match_path(
                dirpath,
                path,
                include_spec,
                global_exclude_spec,
                builtin_exclude_spec,
                user_exclude_spec,
                nested_excludes,
                is_path=False,
            ):
                yield path


def match_path(
    dirpath: Path,
    p: Path,
    include_spec: pathspec.GitIgnoreSpec,
    global_exclude_spec: pathspec.GitIgnoreSpec,
    builtin_exclude_spec: pathspec.GitIgnoreSpec,
    user_exclude_spec: pathspec.GitIgnoreSpec,
    nested_excludes: dict[Path, pathspec.GitIgnoreSpec],
    *,
    is_path: bool,
) -> bool:
    ptype = "directory" if is_path else "file"

    # Always include something included
    if (c := include_spec.check_file(p)).include:
        logger.debug(
            "Including {} {} because it is explicitly included by rule {!r}.",
            ptype,
            p,
            include_spec.patterns[c.index].pattern,  # type: ignore[attr-defined, index]
        )
        return True

    # Always exclude something excluded
    if (c := user_exclude_spec.check_file(p)).include:
        logger.debug(
            "Excluding {} {} because it is explicitly excluded by the user with {!r}.",
            ptype,
            p,
            user_exclude_spec.patterns[c.index].pattern,  # type: ignore[attr-defined, index]
        )
        return False

    # Ignore from global ignore
    if (c := global_exclude_spec.check_file(p)).include:
        logger.debug(
            "Excluding {} {} because it is explicitly excluded by the global ignore with {!r}.",
            ptype,
            p,
            global_exclude_spec.patterns[c.index].pattern,  # type: ignore[attr-defined, index]
        )
        return False

    # Ignore built-in patterns
    if (c := builtin_exclude_spec.check_file(p)).include:
        logger.debug(
            "Excluding {} {} because it is explicitly excluded by the built-in ignore with {!r}.",
            ptype,
            p,
            builtin_exclude_spec.patterns[c.index].pattern,  # type: ignore[attr-defined, index]
        )
        return False

    # Check relative ignores (Python 3.9's is_relative_to workaround)
    for np, nex in nested_excludes.items():
        if (dirpath == np or np in dirpath.parents) and (
            c := nex.check_file(p.relative_to(np))
        ).include:
            logger.debug(
                "Excluding {} {} because it is explicitly excluded by nested ignore with {!r}.",
                ptype,
                p,
                nex.patterns[c.index].pattern,  # type: ignore[attr-defined, index]
            )
            return False

    logger.debug(
        "Including {} {} because it exists (and isn't matched any other way).", ptype, p
    )
    return True



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from collections.abc import Callable


__all__: list[str] = ["_ALL_FIELDS", "_process_dynamic_metadata"]


# Name is not dynamically settable, so not in this list
_STR_FIELDS = frozenset(
    [
        "version",
        "description",
        "requires-python",
        "license",
    ]
)

# Dynamic is not dynamically settable, so not in this list
_LIST_STR_FIELDS = frozenset(
    [
        "classifiers",
        "keywords",
        "dependencies",
        "license-files",
    ]
)

_DICT_STR_FIELDS = frozenset(
    [
        "urls",
        "scripts",
        "gui-scripts",
    ]
)

_LIST_DICT_FIELDS = frozenset(
    [
        "authors",
        "maintainers",
    ]
)

# "dynamic" and "name" can't be set or requested
_ALL_FIELDS = (
    _STR_FIELDS
    | _LIST_STR_FIELDS
    | _DICT_STR_FIELDS
    | _LIST_DICT_FIELDS
    | frozenset(
        [
            "optional-dependencies",
            "readme",
            "entry-points",
        ]
    )
)

T = typing.TypeVar(
    "T",
    bound="str | list[str] | list[dict[str, str]] | dict[str, str] | dict[str, list[str]] | dict[str, dict[str, str]]",
)


def _process_dynamic_metadata(field: str, action: Callable[[str], str], result: T) -> T:
    """
    Helper function for processing an action on the various possible metadata fields.
    """

    if field in _STR_FIELDS:
        if not isinstance(result, str):
            msg = f"Field {field!r} must be a string"
            raise RuntimeError(msg)
        return action(result)  # type: ignore[return-value]
    if field in _LIST_STR_FIELDS:
        if not (isinstance(result, list) and all(isinstance(r, str) for r in result)):
            msg = f"Field {field!r} must be a list of strings"
            raise RuntimeError(msg)
        return [action(r) for r in result]  # type: ignore[arg-type, return-value]
    if field in _DICT_STR_FIELDS | {"readme"}:
        if not isinstance(result, dict) or not all(
            isinstance(v, str) for v in result.values()
        ):
            msg = f"Field {field!r} must be a dictionary of strings"
            raise RuntimeError(msg)
        return {action(k): action(v) for k, v in result.items()}  # type: ignore[arg-type, return-value]
    if field in _LIST_DICT_FIELDS:
        if not isinstance(result, list) or not all(
            isinstance(k, str) and isinstance(v, str)
            for d in result
            for k, v in d.items()  # type: ignore[union-attr]
        ):
            msg = f"Field {field!r} must be a dictionary of strings"
            raise RuntimeError(msg)
        return [{k: action(v) for k, v in d.items()} for d in result]  # type: ignore[union-attr, return-value]
    if field == "entry-points":
        if not isinstance(result, dict) or not all(
            isinstance(d, dict)
            and all(isinstance(k, str) and isinstance(v, str) for k, v in d.items())  # type: ignore[redundant-expr]
            for d in result.values()
        ):
            msg = "Field 'entry-points' must be a dictionary of dictionary of strings"
            raise RuntimeError(msg)
        return {
            dk: {action(k): action(v) for k, v in dv.items()}  # type: ignore[union-attr]
            for dk, dv in result.items()
        }  # type: ignore[return-value]
    if field == "optional-dependencies":
        if not isinstance(result, dict) or not all(
            isinstance(v, list) for v in result.values()
        ):
            msg = "Field 'optional-dependencies' must be a dictionary of lists"
            raise RuntimeError(msg)
        return {k: [action(r) for r in v] for k, v in result.items()}  # type: ignore[return-value]

    msg = f"Unsupported field {field!r} for action"
    raise RuntimeError(msg)
